/*
 * mp3-uniencode
 *  ID3 tags' text encoding converter
 *
 * Written by 2510 on May 22 2007.
 * Copyright by 2510. All rights reserved.
 *
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <iconv.h>

typedef struct id3v2_tag_t id3v2_tag_t;
typedef struct id3v2_t id3v2_t;

struct id3v2_tag_t {
 char name[4];
 int length;
 int flags;
 void *data;
 
 id3v2_tag_t *next;
};

struct id3v2_t {
 unsigned char version1; // id3v2 major-version
 unsigned char version2; // id3v2 minor-version
 unsigned char flags;
 int size;
 
 int modified; // is modified after parsing
 id3v2_tag_t *tag, *last_tag;
};

int option_ascii2latin1 = 0;
const char *source_charset = "CP932";

#pragma mark -

const unsigned char table_itunes1[256] = {
    0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,
    0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,
    0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,
    0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,
    0x40,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,
    0x50,0x51,0x52,0x53,0x54,0x55,0x56,0x57,0x58,0x59,0x5A,0x5B,0x5C,0x5D,0x5E,0x5F,
    0x60,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6A,0x6B,0x6C,0x6D,0x6E,0x6F,
    0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7A,0x7B,0x7C,0x7D,0x7E,0x7F,
    0xC4,0xC5,0xC7,0xC9,0xD1,0xD6,0xDC,0xE1,0xE0,0xE2,0xE4,0xE3,0xE5,0xE7,0xE9,0xE8,
    0xEA,0xEB,0xED,0xEC,0xEE,0xEF,0xF1,0xF3,0xF2,0xF4,0xF6,0xF5,0xFA,0xF9,0xFB,0xFC,
    0xDD,0xB0,0xA2,0xA3,0xA7,0x80,0xB6,0xDF,0xAE,0xA9,0x81,0xB4,0xA8,0x82,0xC6,0xD8,
    0x83,0xB1,0xBE,0x84,0xA5,0xB5,0x8F,0x85,0xBD,0xBC,0x86,0xAA,0xBA,0x87,0xE6,0xF8,
    0xBF,0xA1,0xAC,0x88,0x9F,0x89,0x90,0xAB,0xBB,0x8A,0xA0,0xC0,0xC3,0xD5,0x91,0xA6,
    0xAD,0x8B,0xB3,0xB2,0x8C,0xB9,0xF7,0xD7,0xFF,0x8D,0x8E,0xA4,0xD0,0xF0,0xDE,0xFE,
    0xFD,0xB7,0x92,0x93,0x94,0xC2,0xCA,0xC1,0xCB,0xC8,0xCD,0xCE,0xCF,0xCC,0xD3,0xD4,
    0x95,0xD2,0xDA,0xDB,0xD9,0x9E,0x96,0x97,0xAF,0x98,0x99,0x9A,0xB8,0x9B,0x9C,0x9D
    };
unsigned char table_itunes2[256];

id3v2_t *id3v2_parse(int fd);
int id3v2_store(id3v2_t *self, int fd);
void id3v2_dealloc(id3v2_t *self);

#pragma mark -

int tell(int fd) {
 int retval = (int)lseek(fd, 0, SEEK_CUR);
 if(retval < 0) {
  printf("tell: error '%s'\n", strerror(errno));
  abort();
 }
 return retval;
}

#pragma mark -

int id3v2_parse_v22(id3v2_t *self, int fd, int size) {
 while(size >= 6) {
  unsigned char temp[6];
  id3v2_tag_t *tag;
  
  if(read(fd, temp, 6) != 6) {
   printf("id3v2_parse_v22: failed reading at %d.\n", tell(fd));
   return -1;
  }
  if(memcmp(temp, "\0\0\0", 3) == 0)
   return 0;
  
  tag = (id3v2_tag_t *)malloc(sizeof(id3v2_tag_t));
  memcpy(tag->name, temp, 3);
  tag->name[3] = 0;
  tag->length = temp[3] << 16 | temp[4] << 8 | temp[5];
  tag->flags = 0;
  tag->next = 0;
  
  tag->data = malloc(tag->length);
  if(read(fd, tag->data, tag->length) != tag->length) {
   printf("id3v2_parse_v22: failed reading at %d.\n", tell(fd));
   free(tag->data);
   free(tag);
   return -1;
  }
  
  /* printf("id3v2_parse_v22: found tag '%c%c%c', %d bytes.\n",
   tag->name[0], tag->name[1], tag->name[2], tag->length); */
  
  if(self->tag)
   self->last_tag->next = tag;
  else
   self->tag = tag;
  self->last_tag = tag;
  
  size -= 6 + tag->length;
 }
 return 0;
}

int id3v2_parse_v23(id3v2_t *self, int fd, int size) {
 while(size >= 10) {
  unsigned char temp[10];
  id3v2_tag_t *tag;
  
  if(read(fd, temp, 10) != 10) {
   printf("id3v2_parse_v23: failed reading at %d.\n", tell(fd));
   return -1;
  }
  if(memcmp(temp, "\0\0\0\0", 4) == 0)
   return 0;
  
  tag = (id3v2_tag_t *)malloc(sizeof(id3v2_tag_t));
  memcpy(tag->name, temp, 4);
  tag->length = temp[4] << 24 | temp[5] << 16 | temp[6] << 8 | temp[7];
  tag->flags = temp[8] << 8 | temp[9];
  tag->next = 0;
  
  tag->data = malloc(tag->length);
  if(read(fd, tag->data, tag->length) != tag->length) {
   printf("id3v2_parse_v23: failed reading at %d.\n", tell(fd));
   free(tag->data);
   free(tag);
   return -1;
  }
  
  /* printf("id3v2_parse_v23: found tag '%c%c%c%c', flags %04X, %d bytes.\n",
   tag->name[0], tag->name[1], tag->name[2], tag->name[3],
   tag->flags, tag->length); */
  
  if(self->tag)
   self->last_tag->next = tag;
  else
   self->tag = tag;
  self->last_tag = tag;
  
  size -= 10 + tag->length;
 }
 return 0;
}

id3v2_t *id3v2_parse(int fd) {
 unsigned char temp[10];
 id3v2_t *self = 0;
 int endpos;
 
 do {
  lseek(fd, 0, SEEK_SET);
  if(read(fd, temp, 10) != 10) {
   printf("id3v2_parse: failed reading header.\n");
   break;
  }
  if(memcmp(temp, "ID3", 3) != 0) {
   printf("id3v2_parse: ID3v2 header not found.\n");
   break;
  }
  self = (id3v2_t *)malloc(sizeof(id3v2_t));
  self->version1 = temp[3];
  self->version2 = temp[4];
  self->flags = temp[5];
  self->size = temp[6] << 21 | temp[7] << 14 | temp[8] << 7 | temp[9];
  /* printf("id3v2_parse: found ID3v2, version 2.%d.%d, size %d bytes.\n", self->version1, self->version2, self->size); */
  
  endpos = tell(fd) + self->size;
  
  if(self->flags & 0x40) {
   printf("id3v2_parse: compression is not supported.\n");
   break;
  }
  if(self->flags & 0x80) {
   printf("id3v2_parse: synchronisation scheme is not supported.\n");
   break;
  }
  
  if(self->version1 == 2) {
   if(id3v2_parse_v22(self, fd, self->size))
    break;
  } else if(self->version1 == 3 || self->version1 == 4) {
   if(id3v2_parse_v23(self, fd, self->size))
    break;
  } else {
   printf("id3v2_parse: unsupported version.\n");
   break;
  }
  
  /* printf("id3v2_parse: trailing dummy data is %d bytes.\n", endpos - tell(fd)); */
  lseek(fd, endpos, SEEK_SET);
  self->modified = 0;
  return self;
 } while(0);
 
 if(self)
  id3v2_dealloc(self);
 return 0;
}

#pragma mark -

int id3v2_store_v22(id3v2_t *self, int fd) {
 id3v2_tag_t *tag;
 
 tag = self->tag;
 while(tag) {
  unsigned char temp[6];
  memcpy(temp, tag->name, 3);
  temp[3] = tag->length >> 16;
  temp[4] = tag->length >> 8;
  temp[5] = tag->length >> 0;
  
  /* printf("id3v2_store_v22: storing tag '%c%c%c', %d bytes.\n",
   tag->name[0], tag->name[1], tag->name[2], tag->length); */
  
  if(write(fd, temp, 6) != 6) {
   printf("id3v2_store_v22: failed writing at %d.\n", tell(fd));
   return -1;
  }
  if(write(fd, tag->data, tag->length) != tag->length) {
   printf("id3v2_store_v22: failed writing at %d.\n", tell(fd));
   return -1;
  }
  
  tag = tag->next;
 }
 return 0;
}

int id3v2_store_v23(id3v2_t *self, int fd) {
 id3v2_tag_t *tag;
 
 tag = self->tag;
 while(tag) {
  unsigned char temp[10];
  memcpy(&temp[0], tag->name, 4);
  temp[4] = tag->length >> 24;
  temp[5] = tag->length >> 16;
  temp[6] = tag->length >> 8;
  temp[7] = tag->length >> 0;
  temp[8] = tag->flags >> 8;
  temp[9] = tag->flags >> 0;
  
  /* printf("id3v2_store_v23: storing tag '%c%c%c%c', flags %04X, %d bytes.\n",
   tag->name[0], tag->name[1], tag->name[2], tag->name[3],
   tag->flags, tag->length); */
  
  if(write(fd, temp, 10) != 10) {
   printf("id3v2_store_v23: failed writing at %d.\n", tell(fd));
   return -1;
  }
  if(write(fd, tag->data, tag->length) != tag->length) {
   printf("id3v2_store_v23: failed writing at %d.\n", tell(fd));
   return -1;
  }
  
  tag = tag->next;
 }
 return 0;
}

int id3v2_store(id3v2_t *self, int fd) {
 unsigned char temp[10];
 unsigned char empty[64];
 int lastpos, length, padding_length;
 
 memcpy(&temp[0], "ID3", 3);
 temp[3] = self->version1;
 temp[4] = self->version2;
 temp[5] = self->flags;
 temp[6] = 0;
 temp[7] = 0;
 temp[8] = 0;
 temp[9] = 0;
 
 if(write(fd, temp, 10) != 10) {
  printf("id3v2_store: failed writing header.\n");
  return -1;
 }
 
 /* printf("id3v2_store: version 2.%d.%d\n", self->version1, self->version2); */
 
 lastpos = tell(fd);
 if(self->version1 == 2) {
  if(id3v2_store_v22(self, fd))
   return -1;
 } else if(self->version1 == 3 || self->version1 == 4) {
  if(id3v2_store_v23(self, fd))
   return -1;
 }
 length = tell(fd) - lastpos;
 
 memset(empty, 0x00, sizeof(empty));
 padding_length = 16;
 if(length & 15)
  padding_length += 16 - (length & 15);
 if(write(fd, empty, padding_length) != padding_length) {
  printf("id3v2_store: failed writing padding area.\n");
  return -1;
 }
 /* printf("id3v2_store: data length %d + padding %d bytes\n", length, padding_length); */
 length += padding_length;
 
 temp[6] = (length >> 21) & 0x7F;
 temp[7] = (length >> 14) & 0x7F;
 temp[8] = (length >> 7) & 0x7F;
 temp[9] = (length >> 0) & 0x7F;
 
 lseek(fd, -(10 + length), SEEK_CUR);
 if(write(fd, temp, 10) != 10) {
  printf("id3v2_store: failed writing header.\n");
  return -1;
 }
 lseek(fd, length, SEEK_CUR);
 return 0;
}

#pragma mark -

void id3v2_dealloc(id3v2_t *self) {
 id3v2_tag_t *tag;
 if(!self)
  return;
 tag = self->tag;
 while(tag) {
  id3v2_tag_t *next = tag->next;
  free(tag->data);
  free(tag);
  tag = next;
 }
 free(self);
}

#pragma mark -

char *id3v2_itunes_ascii2latin1(const char *string) {
 int pos, length = strlen(string + 1);
 char *decoded = (char *)malloc(1 + length + 1);
 
 decoded[0] = string[0];
 for(pos = 0;pos <= length;pos ++)
  decoded[1 + pos] = table_itunes2[((int)string[1 + pos]) & 0xFF];
 return decoded;
}

void id3v2_uniencode_string(const char *string, int in_length, char **unicoded, int *out_length) {
 int failure = 0;
 char *outstring, *outptr;
 size_t insize = in_length - 1;
 size_t outsize, outspace;
 iconv_t converter = iconv_open("UTF-16", source_charset);
 
 *unicoded = 0;
 string ++;
 
 outsize = outspace = 256;
 outptr = outstring = (char *)malloc(outsize);
 *outptr ++ = 0x01;
 outspace --;
 
 while(!failure) {
  if(insize == 0)
   if(string == 0)
    break;
   else
    string = 0;
  while(iconv(converter, &string, &insize, &outptr, &outspace) == (size_t)-1) {
   if(errno == E2BIG) {
    size_t offset = outptr - outstring;
    outsize *= 2;
    outstring = realloc(outstring, outsize + 1);
    outptr = outstring + offset;
    outspace = outsize - offset;
   } else if(errno == EILSEQ) {
    string += 2;
    insize -= 2;
   } else {
    printf("id3v2_uniencode_string: iconv failed (%s)\n", strerror(errno));
    failure = 1;
    break;
   }
  }
 }
 iconv_close(converter);
 
 if(failure) {
  free(outstring);
  return;
 }
 
 outsize = outptr - outstring;
 outstring[outsize ++] = 0;
 outstring[outsize ++] = 0;
 *unicoded = outstring;
 *out_length = outsize;
}


void id3v2_uniencode(id3v2_t *self, const char *tagname) {
 id3v2_tag_t *tag;
 tag = self->tag;
 while(tag) {
  int match = 0;
  if(self->version1 == 2) {
   if(!memcmp(tag->name, tagname, 3))
    match = 1;
  } else if(self->version1 == 3 || self->version1 == 4) {
   if(!memcmp(tag->name, tagname, 4))
    match = 1;
  }
  
  /* is data ASCII formatted? */
  if(((char *)tag->data)[0] != 0x00)
   match = 0;
  else
  {
   char ch, *ptr = tag->data;
   ptr ++;
   while((ch = *ptr ++) != 0)
    if(ch & 0x80)
     break;
   if(ch == 0)
    match = 0;
  }
  
  /* decode iTunes coding */
  if(match && option_ascii2latin1) {
   char *data = tag->data;
   tag->data = id3v2_itunes_ascii2latin1(data);
   if(tag->data)
    free(data);
   else {
    tag->data = data;
    match = 0;
   }
  }
  
  /* convert to unicode */
  if(match) {
   char *data;
   int length;
   id3v2_uniencode_string(tag->data, tag->length, &data, &length);
   if(data) {
    free(tag->data);
    tag->data = data;
    tag->length = length;
    self->modified = 1;
   } else
    match = 0;
  }
  
  /* finished */
  if(match) {
   unsigned char *ptr = tag->data;
   printf("id3v2_uniencode: converted tag '%c%c%c%c', heading %02X %02X %02X %02X %02X\n",
    tag->name[0], tag->name[1], tag->name[2], tag->name[3],
    ptr[0], ptr[1], ptr[2], ptr[3], ptr[4]
    );
  }
  
  tag = tag->next;
 }
}

void uniencode(const char *path) {
 char *bakpath;
 int fd;
 id3v2_t *tag;
 char *mp3_data;
 int mp3_start, mp3_size;
 
 bakpath = (char *)malloc(strlen(path) + 5);
 strcpy(bakpath, path);
 strcat(bakpath, ".bak");
 
 /* parse source file */
 fd = open(path, O_RDONLY);
 do {
  tag = id3v2_parse(fd);
  if(!tag)
   break;
  
  mp3_start = tell(fd);
  mp3_size = lseek(fd, 0, SEEK_END) - mp3_start;
  mp3_data = malloc(mp3_size);
  lseek(fd, mp3_start, SEEK_SET);
  if(read(fd, mp3_data, mp3_size) != mp3_size) {
   printf("main: failed reading mp3 body at %d.\n", tell(fd));
   id3v2_dealloc(tag);
   tag = 0;
   break;
  }
 } while(0);
 close(fd);
 
 if(!tag)
  return;
 
 /* convert to Unicode */
 
 /* v2.2 tags */
 id3v2_uniencode(tag, "TT2\0"); /* title */
 id3v2_uniencode(tag, "TP1\0"); /* artist */
 id3v2_uniencode(tag, "TAL\0"); /* album */
 id3v2_uniencode(tag, "TCM\0"); /* composer */
 id3v2_uniencode(tag, "TCO\0"); /* genre */
 
 /* v2.3 tags */
 id3v2_uniencode(tag, "TIT2"); /* title */
 id3v2_uniencode(tag, "TPE1"); /* artist */
 id3v2_uniencode(tag, "TALB"); /* album */
 id3v2_uniencode(tag, "TCOM"); /* composer */
 id3v2_uniencode(tag, "TCON"); /* genre */
 
 /* */
 /* rename(path, bakpath); */
 
 if(tag->modified) {
  /* store data into a new file */
  fd = open(path, O_CREAT | O_RDONLY | O_WRONLY, 0644);
  id3v2_store(tag, fd);
  if(write(fd, mp3_data, mp3_size) != mp3_size)
   printf("main: failed writing mp3 body at %d.\n", tell(fd));
  close(fd);
 }
 id3v2_dealloc(tag);
 
 free(bakpath);
}

int main(int argc, const char *argv[]) {
 int n, argp;
 
 /* read options */
 for(argp = 1; argp < argc && argv[argp][0] == '-'; argp ++) {
  const char *arg = argv[argp];
  if(!strcmp(arg, "-a")) /* ascii to iso-latin1 conversion */
   option_ascii2latin1 = 1;
  else if(!strcmp(arg, "-c")) { /* source coding set */
   iconv_t converter;
   if(++ argp == argc) {
    printf("mp3-uniencode: source coding set name for -c option is missing.\n");
    return 1;
   }
   converter = iconv_open("UTF-16", argv[argp]);
   if(converter == (iconv_t)-1) {
    printf("mp3-uniencode: cannot use '%s' for source coding set. (%s)\n", argv[argp], strerror(errno));
    return 1;
   }
   iconv_close(converter);
  }
 }
 
 if(argp == argc) {
  puts(
   "mp3-uniencode: no input file specified.\n"
   " usage: mp3-uniencode [options...] <input files...>\n"
   " -a\n"
   "  Do 'ASCII to ISO-Latin1' conversion before transcoding to Unicode.\n"
   " -c <source coding set>\n"
   "  Set source text encoding format name to <source coding set>. Default is CP932."
   );
  return 1;
 }
 
 /* create a reverse conversion table */
 for(n = 0;n < 256;n ++)
  table_itunes2[table_itunes1[n]] = n;
 
 /* process */
 for(n = argp;n < argc;n ++) {
  printf("processing: %s\n", argv[n]);
  uniencode(argv[n]);
 }
 return 0;
}
